@@ -28,11 +28,16 @@ case "${architecture}" in
     ;;
 esac
 
-pgdg_rpm=pgdg-centos${PG_PACKAGE_VERSION}-${PG_VERSION}-3.noarch.rpm
+if [ ${PG_PACKAGE_VERSION} -ge 94 ]; then
+  pgdg_rpm=pgdg-centos${PG_PACKAGE_VERSION}-${PG_VERSION}-3.noarch.rpm
+else
+  pgdg_rpm=pgdg-centos${PG_PACKAGE_VERSION}-${PG_VERSION}-2.noarch.rpm
+fi
 run wget --no-check-certificate https://yum.postgresql.org/${PG_VERSION}/redhat/rhel-${distribution_version}-${architecture}/${pgdg_rpm}
 run rpm -ivh ${pgdg_rpm}
-run rpm -ivh http://packages.groonga.org/centos/groonga-release-1.1.0-1.noarch.rpm
-run yum makecache
+groonga_release_rpm=groonga-release-1.3.0-1.noarch.rpm
+groonga_release_rpm_url=https://packages.groonga.org/centos/${groonga_release_rpm}
+run yum install -y ${groonga_release_rpm_url}
 
 run yum groupinstall -y "Development Tools"
 run yum install -y rpm-build rpmdevtools tar ${DEPENDED_PACKAGES}
@@ -68,6 +73,11 @@ else
 fi
 run cp /vagrant/tmp/${distribution}/${PACKAGE}.spec rpmbuild/SPECS/
 
+if grep -q -E 'Requires:\s+msgpack' rpmbuild/SPECS/${PACKAGE}.spec; then
+  run yum install -y epel-release
+  run yum install -y msgpack-devel
+fi
+
 run rpmbuild -ba ${rpmbuild_options} rpmbuild/SPECS/${PACKAGE}.spec
 
 run mv rpmbuild/RPMS/*/* "${rpm_dir}/"
