@@ -6,6 +6,14 @@
 #include "pgrn-groonga.h"
 #include "pgrn-index-status.h"
 #include "pgrn-wal.h"
+#include "pgrn-writable.h"
+
+#ifdef WIN32
+#	define PRId64 "I64d"
+#	define PRIu64 "I64u"
+#else
+#	include <inttypes.h>
+#endif
 
 static bool PGrnWALEnabled = false;
 
@@ -29,14 +37,24 @@ PGrnWALDisable(void)
 
 #ifdef PGRN_SUPPORT_WAL
 #	include <access/generic_xlog.h>
+#	include <access/heapam.h>
+#	include <access/htup_details.h>
+#	include <miscadmin.h>
 #	include <storage/bufmgr.h>
 #	include <storage/bufpage.h>
 #	include <storage/lmgr.h>
 #	include <storage/lockdefs.h>
+#	include <utils/acl.h>
+#	include <utils/builtins.h>
 
 #	include <msgpack.h>
 #endif
 
+PGRN_FUNCTION_INFO_V1(pgroonga_wal_apply_index);
+PGRN_FUNCTION_INFO_V1(pgroonga_wal_apply_all);
+PGRN_FUNCTION_INFO_V1(pgroonga_wal_truncate_index);
+PGRN_FUNCTION_INFO_V1(pgroonga_wal_truncate_all);
+
 #ifdef PGRN_SUPPORT_WAL
 static grn_ctx *ctx = &PGrnContext;
 static struct PGrnBuffers *buffers = &PGrnBuffers;
@@ -47,7 +65,9 @@ typedef enum {
 	PGRN_WAL_ACTION_INSERT,
 	PGRN_WAL_ACTION_CREATE_TABLE,
 	PGRN_WAL_ACTION_CREATE_COLUMN,
-	PGRN_WAL_ACTION_SET_SOURCES
+	PGRN_WAL_ACTION_SET_SOURCES,
+	PGRN_WAL_ACTION_RENAME_TABLE,
+	PGRN_WAL_ACTION_DELETE
 } PGrnWALAction;
 
 #define PGRN_WAL_META_PAGE_SPECIAL_VERSION 1
@@ -166,6 +186,24 @@ msgpack_pack_grn_obj(msgpack_packer *packer, grn_obj *object)
 #define PGRN_WAL_META_PAGE_BLOCK_NUMBER 0
 
 #ifdef PGRN_SUPPORT_WAL
+static Buffer
+PGrnWALReadLockedBuffer(Relation index,
+						BlockNumber blockNumber,
+						int bufferLockMode)
+{
+	LOCKMODE lockMode = ExclusiveLock;
+	Buffer buffer;
+
+	if (blockNumber == P_NEW)
+		LockRelationForExtension(index, lockMode);
+	buffer = ReadBuffer(index, blockNumber);
+	LockBuffer(buffer, bufferLockMode);
+	if (blockNumber == P_NEW)
+		UnlockRelationForExtension(index, lockMode);
+
+	return buffer;
+}
+
 static void
 PGrnWALDataInitNUsedPages(PGrnWALData *data)
 {
@@ -177,11 +215,10 @@ PGrnWALDataInitMeta(PGrnWALData *data)
 {
 	if (RelationGetNumberOfBlocks(data->index) == 0)
 	{
-		LockRelationForExtension(data->index, ExclusiveLock);
-		data->meta.buffer = ReadBuffer(data->index, P_NEW);
-		LockBuffer(data->meta.buffer, BUFFER_LOCK_EXCLUSIVE);
-		UnlockRelationForExtension(data->index, ExclusiveLock);
-
+		data->meta.buffer =
+			PGrnWALReadLockedBuffer(data->index,
+									P_NEW,
+									BUFFER_LOCK_EXCLUSIVE);
 		data->meta.page = GenericXLogRegisterBuffer(data->state,
 													data->meta.buffer,
 													GENERIC_XLOG_FULL_IMAGE);
@@ -194,10 +231,10 @@ PGrnWALDataInitMeta(PGrnWALData *data)
 	}
 	else
 	{
-		data->meta.buffer = ReadBuffer(data->index,
-									   PGRN_WAL_META_PAGE_BLOCK_NUMBER);
-		LockBuffer(data->meta.buffer, BUFFER_LOCK_EXCLUSIVE);
-
+		data->meta.buffer =
+			PGrnWALReadLockedBuffer(data->index,
+									PGRN_WAL_META_PAGE_BLOCK_NUMBER,
+									BUFFER_LOCK_EXCLUSIVE);
 		data->meta.page = GenericXLogRegisterBuffer(data->state,
 													data->meta.buffer,
 													0);
@@ -283,13 +320,11 @@ PGrnWALPageWriterEnsureCurrent(PGrnWALData *data)
 	meta = data->meta.pageSpecial;
 	if (RelationGetNumberOfBlocks(data->index) <= meta->next)
 	{
-		LockRelationForExtension(data->index, ExclusiveLock);
-		data->current.buffer = ReadBuffer(data->index, P_NEW);
-		LockBuffer(data->current.buffer, BUFFER_LOCK_EXCLUSIVE);
-		UnlockRelationForExtension(data->index, ExclusiveLock);
-
+		data->current.buffer =
+			PGrnWALReadLockedBuffer(data->index,
+									P_NEW,
+									BUFFER_LOCK_EXCLUSIVE);
 		meta->next = BufferGetBlockNumber(data->current.buffer);
-
 		data->current.page =
 			GenericXLogRegisterBuffer(data->state,
 									  data->current.buffer,
@@ -298,9 +333,10 @@ PGrnWALPageWriterEnsureCurrent(PGrnWALData *data)
 	}
 	else
 	{
-		data->current.buffer = ReadBuffer(data->index, meta->next);
-		LockBuffer(data->current.buffer, BUFFER_LOCK_EXCLUSIVE);
-
+		data->current.buffer =
+			PGrnWALReadLockedBuffer(data->index,
+									meta->next,
+									BUFFER_LOCK_EXCLUSIVE);
 		data->current.page =
 			GenericXLogRegisterBuffer(data->state,
 									  data->current.buffer,
@@ -537,7 +573,7 @@ PGrnWALInsertColumnValueRaw(PGrnWALData *data,
 		msgpack_pack_uint64(packer, *((uint64_t *)(value)));
 		break;
 	case GRN_DB_FLOAT:
-		msgpack_pack_double(packer, *((float_t *)(value)));
+		msgpack_pack_double(packer, *((double *)(value)));
 		break;
 	case GRN_DB_TIME:
 		msgpack_pack_int64(packer, *((int64_t *)(value)));
@@ -663,9 +699,6 @@ PGrnWALInsertColumn(PGrnWALData *data,
 
 	nameSize = grn_column_name(ctx, column, name, GRN_TABLE_MAX_KEY_SIZE);
 
-	if (!data)
-		return;
-
 	PGrnWALInsertColumnStart(data, name, nameSize);
 
 	switch (value->header.type)
@@ -870,6 +903,72 @@ PGrnWALSetSourceIDs(Relation index,
 #endif
 }
 
+void
+PGrnWALRenameTable(Relation index,
+				   const char *name,
+				   size_t nameSize,
+				   const char *newName,
+				   size_t newNameSize)
+{
+#ifdef PGRN_SUPPORT_WAL
+	PGrnWALData *data;
+	msgpack_packer *packer;
+	size_t nElements = 3;
+
+	data = PGrnWALStart(index);
+	if (!data)
+		return;
+
+	packer = &(data->packer);
+	msgpack_pack_map(packer, nElements);
+
+	msgpack_pack_cstr(packer, "_action");
+	msgpack_pack_uint32(packer, PGRN_WAL_ACTION_RENAME_TABLE);
+
+	msgpack_pack_cstr(packer, "name");
+	msgpack_pack_str(packer, nameSize);
+	msgpack_pack_str_body(packer, name, nameSize);
+
+	msgpack_pack_cstr(packer, "new_name");
+	msgpack_pack_str(packer, newNameSize);
+	msgpack_pack_str_body(packer, newName, newNameSize);
+
+	PGrnWALFinish(data);
+#endif
+}
+
+void
+PGrnWALDelete(Relation index,
+			  grn_obj *table,
+			  const char *key,
+			  size_t keySize)
+{
+#ifdef PGRN_SUPPORT_WAL
+	PGrnWALData *data;
+	msgpack_packer *packer;
+	size_t nElements = 3;
+
+	data = PGrnWALStart(index);
+	if (!data)
+		return;
+
+	packer = &(data->packer);
+	msgpack_pack_map(packer, nElements);
+
+	msgpack_pack_cstr(packer, "_action");
+	msgpack_pack_uint32(packer, PGRN_WAL_ACTION_DELETE);
+
+	msgpack_pack_cstr(packer, "table");
+	msgpack_pack_grn_obj(packer, table);
+
+	msgpack_pack_cstr(packer, "key");
+	msgpack_pack_bin(packer, keySize);
+	msgpack_pack_bin_body(packer, key, keySize);
+
+	PGrnWALFinish(data);
+#endif
+}
+
 #ifdef PGRN_SUPPORT_WAL
 typedef struct {
 	Relation index;
@@ -893,6 +992,8 @@ PGrnWALApplyNeeded(PGrnWALApplyData *data)
 	PGrnIndexStatusGetWALAppliedPosition(data->index,
 										 &currentBlock,
 										 &currentOffset);
+	if (currentBlock == PGRN_WAL_META_PAGE_BLOCK_NUMBER)
+		currentBlock++;
 
 	nBlocks = RelationGetNumberOfBlocks(data->index);
 	if (currentBlock >= nBlocks)
@@ -905,17 +1006,32 @@ PGrnWALApplyNeeded(PGrnWALApplyData *data)
 		Page page;
 		bool needToApply;
 
-		buffer = ReadBuffer(data->index, currentBlock);
-		LockBuffer(buffer, BUFFER_LOCK_SHARE);
+		buffer = PGrnWALReadLockedBuffer(data->index,
+										 currentBlock,
+										 BUFFER_LOCK_SHARE);
 		page = BufferGetPage(buffer);
 		needToApply = (PGrnWALPageGetLastOffset(page) > currentOffset);
 		UnlockReleaseBuffer(buffer);
-		return needToApply;
+		if (!needToApply)
+			return false;
 	}
 	else
 	{
-		return true;
+		Buffer buffer;
+		Page page;
+		bool needToApply;
+
+		buffer = PGrnWALReadLockedBuffer(data->index,
+										 currentBlock,
+										 BUFFER_LOCK_SHARE);
+		page = BufferGetPage(buffer);
+		needToApply = (PGrnWALPageGetLastOffset(page) > 0);
+		UnlockReleaseBuffer(buffer);
+		if (!needToApply)
+			return false;
 	}
+
+	return PGrnIsWritable();
 }
 
 static bool
@@ -990,6 +1106,30 @@ PGrnWALApplyValueGetString(const char *context,
 	*stringSize = MSGPACK_OBJECT_VIA_STR(kv->val).size;
 }
 
+static void
+PGrnWALApplyValueGetBinary(const char *context,
+						   msgpack_object_kv *kv,
+						   const char **binary,
+						   size_t *binarySize)
+{
+	if (kv->val.type != MSGPACK_OBJECT_BIN)
+	{
+		ereport(ERROR,
+				(errcode(ERRCODE_INVALID_PARAMETER_VALUE),
+				 errmsg("pgroonga: WAL: apply: %s%s"
+						"%.*s value must be binary: "
+						"<%#x>",
+						context ? context : "",
+						context ? ": " : "",
+						MSGPACK_OBJECT_VIA_BIN(kv->key).size,
+						MSGPACK_OBJECT_VIA_BIN(kv->key).ptr,
+						kv->val.type)));
+	}
+
+	*binary = MSGPACK_OBJECT_VIA_BIN(kv->val).ptr;
+	*binarySize = MSGPACK_OBJECT_VIA_BIN(kv->val).size;
+}
+
 static grn_obj *
 PGrnWALApplyValueGetGroongaObject(const char *context,
 								  msgpack_object_kv *kv)
@@ -1179,6 +1319,23 @@ PGrnWALApplyInsertArray(PGrnWALApplyData *data,
 			case GRN_DB_UINT64:
 				GRN_UINT64_PUT(ctx, value, ELEMENT_VALUE);
 				break;
+			default:
+				{
+					grn_obj key;
+					if (element->type == MSGPACK_OBJECT_POSITIVE_INTEGER)
+					{
+						GRN_INT64_INIT(&key, 0);
+						GRN_INT64_SET(ctx, &key, element->via.i64);
+					}
+					else
+					{
+						GRN_UINT64_INIT(&key, 0);
+						GRN_UINT64_SET(ctx, &key, element->via.u64);
+					}
+					grn_obj_cast(ctx, &key, value, GRN_FALSE);
+					GRN_OBJ_FIN(ctx, &key);
+				}
+				break;
 			}
 			break;
 #undef ELEMENT_VALUE
@@ -1242,19 +1399,7 @@ PGrnWALApplyInsert(PGrnWALApplyData *data,
 		kv = &(map->ptr[currentElement]);
 		if (PGrnWALApplyKeyEqual(context, &(kv->key), GRN_COLUMN_NAME_KEY))
 		{
-			if (kv->val.type != MSGPACK_OBJECT_BIN)
-			{
-				ereport(ERROR,
-						(errcode(ERRCODE_INVALID_PARAMETER_VALUE),
-						 errmsg("pgroonga: WAL: apply: %s: "
-								"%s value must be binary: "
-								"<%#x>",
-								context,
-								GRN_COLUMN_NAME_KEY,
-								kv->val.type)));
-			}
-			key = MSGPACK_OBJECT_VIA_BIN(kv->val).ptr;
-			keySize = MSGPACK_OBJECT_VIA_BIN(kv->val).size;
+			PGrnWALApplyValueGetBinary(context, kv, &key, &keySize);
 			currentElement++;
 		}
 	}
@@ -1460,6 +1605,87 @@ PGrnWALApplySetSources(PGrnWALApplyData *data,
 	grn_obj_set_info(ctx, column, GRN_INFO_SOURCE, sourceIDs);
 }
 
+static void
+PGrnWALApplyRenameTable(PGrnWALApplyData *data,
+						msgpack_object_map *map,
+						uint32_t currentElement)
+{
+	const char *context = "rename table";
+	grn_obj *table = NULL;
+	const char *newName = NULL;
+	size_t newNameSize = 0;
+	uint32_t i;
+
+	for (i = currentElement; i < map->size; i++)
+	{
+		msgpack_object_kv *kv;
+
+		kv = &(map->ptr[i]);
+		if (PGrnWALApplyKeyEqual(context, &(kv->key), "name"))
+		{
+			table = PGrnWALApplyValueGetGroongaObject(context, kv);
+		}
+		else if (PGrnWALApplyKeyEqual(context, &(kv->key), "new_name"))
+		{
+			PGrnWALApplyValueGetString(context, kv, &newName, &newNameSize);
+		}
+	}
+
+	grn_table_rename(ctx, table, newName, newNameSize);
+}
+
+static void
+PGrnWALApplyDelete(PGrnWALApplyData *data,
+				   msgpack_object_map *map,
+				   uint32_t currentElement)
+{
+	const char *context = "delete";
+	grn_obj *table = NULL;
+	const char *key = NULL;
+	size_t keySize = 0;
+	uint32_t i;
+
+	for (i = currentElement; i < map->size; i++)
+	{
+		msgpack_object_kv *kv;
+
+		kv = &(map->ptr[i]);
+		if (PGrnWALApplyKeyEqual(context, &(kv->key), "table"))
+		{
+			table = PGrnWALApplyValueGetGroongaObject(context, kv);
+		}
+		else if (PGrnWALApplyKeyEqual(context, &(kv->key), "key"))
+		{
+			PGrnWALApplyValueGetBinary(context, kv, &key, &keySize);
+			currentElement++;
+		}
+	}
+
+	if (table->header.type == GRN_TABLE_NO_KEY)
+	{
+		const uint64 packedCtid = *((uint64 *)key);
+		grn_obj *ctidColumn = grn_obj_column(ctx, table, "ctid", strlen("ctid"));
+		grn_obj ctidValue;
+		GRN_UINT64_INIT(&ctidValue, 0);
+		GRN_TABLE_EACH_BEGIN(ctx, table, cursor, id)
+		{
+			GRN_BULK_REWIND(&ctidValue);
+			grn_obj_get_value(ctx, ctidColumn, id, &ctidValue);
+			if (packedCtid == GRN_UINT64_VALUE(&ctidValue))
+			{
+				grn_table_cursor_delete(ctx, cursor);
+				break;
+			}
+		} GRN_TABLE_EACH_END(ctx, cursor);
+		GRN_OBJ_FIN(ctx, &ctidValue);
+		grn_obj_unlink(ctx, ctidColumn);
+	}
+	else
+	{
+		grn_table_delete(ctx, table, key, keySize);
+	}
+}
+
 static void
 PGrnWALApplyObject(PGrnWALApplyData *data, msgpack_object *object)
 {
@@ -1470,10 +1696,100 @@ PGrnWALApplyObject(PGrnWALApplyData *data, msgpack_object *object)
 
 	if (object->type != MSGPACK_OBJECT_MAP)
 	{
-		ereport(ERROR,
-				(errcode(ERRCODE_INVALID_PARAMETER_VALUE),
-				 errmsg("pgroonga: WAL: apply: record must be map: <%#x>",
-						object->type)));
+		const char *message =
+			"pgroonga: WAL: apply: record must be map";
+		BlockNumber currentBlock;
+		LocationIndex currentOffset;
+
+		PGrnIndexStatusGetWALAppliedPosition(data->index,
+											 &currentBlock,
+											 &currentOffset);
+		switch (object->type)
+		{
+		case MSGPACK_OBJECT_NIL:
+			ereport(ERROR,
+					(errcode(ERRCODE_INVALID_PARAMETER_VALUE),
+					 errmsg("%s: <%u><%u>: <nil>",
+							message,
+							currentBlock,
+							currentOffset)));
+			break;
+		case MSGPACK_OBJECT_BOOLEAN:
+			ereport(ERROR,
+					(errcode(ERRCODE_INVALID_PARAMETER_VALUE),
+					 errmsg("%s: <%u><%u>: <boolean>: <%s>",
+							message,
+							currentBlock,
+							currentOffset,
+							object->via.boolean ? "true" : "false")));
+			break;
+		case MSGPACK_OBJECT_POSITIVE_INTEGER:
+			ereport(ERROR,
+					(errcode(ERRCODE_INVALID_PARAMETER_VALUE),
+					 errmsg("%s: <%u><%u>: <positive-integer>: <%" PRId64 ">",
+							message,
+							currentBlock,
+							currentOffset,
+							object->via.i64)));
+			break;
+		case MSGPACK_OBJECT_NEGATIVE_INTEGER:
+			ereport(ERROR,
+					(errcode(ERRCODE_INVALID_PARAMETER_VALUE),
+					 errmsg("%s: <%u><%u>: <negative-integer>: <%" PRIu64 ">",
+							message,
+							currentBlock,
+							currentOffset,
+							object->via.u64)));
+			break;
+		case MSGPACK_OBJECT_FLOAT:
+			ereport(ERROR,
+					(errcode(ERRCODE_INVALID_PARAMETER_VALUE),
+					 errmsg("%s: <%u><%u>: <float>: <%g>",
+							message,
+							currentBlock,
+							currentOffset,
+							MSGPACK_OBJECT_VIA_FLOAT(*object))));
+			break;
+		case MSGPACK_OBJECT_STR:
+			ereport(ERROR,
+					(errcode(ERRCODE_INVALID_PARAMETER_VALUE),
+					 errmsg("%s: <%u><%u>: <string>: <%.*s>",
+							message,
+							currentBlock,
+							currentOffset,
+							(int) MSGPACK_OBJECT_VIA_STR(*object).size,
+							MSGPACK_OBJECT_VIA_STR(*object).ptr)));
+			break;
+		case MSGPACK_OBJECT_ARRAY:
+			ereport(ERROR,
+					(errcode(ERRCODE_INVALID_PARAMETER_VALUE),
+					 errmsg("%s: <%u><%u>: <array>: <%u>",
+							message,
+							currentBlock,
+							currentOffset,
+							object->via.array.size)));
+			break;
+#if MSGPACK_VERSION_MAJOR != 0
+		case MSGPACK_OBJECT_BIN:
+			ereport(ERROR,
+					(errcode(ERRCODE_INVALID_PARAMETER_VALUE),
+					 errmsg("%s: <%u><%u>: <binary>: <%u>",
+							message,
+							currentBlock,
+							currentOffset,
+							MSGPACK_OBJECT_VIA_BIN(*object).size)));
+			break;
+#endif
+		default:
+			ereport(ERROR,
+					(errcode(ERRCODE_INVALID_PARAMETER_VALUE),
+					 errmsg("%s: <%u><%u>: <%#x>",
+							message,
+							currentBlock,
+							currentOffset,
+							object->type)));
+			break;
+		}
 	}
 
 	map = &(object->via.map);
@@ -1504,6 +1820,12 @@ PGrnWALApplyObject(PGrnWALApplyData *data, msgpack_object *object)
 	case PGRN_WAL_ACTION_SET_SOURCES:
 		PGrnWALApplySetSources(data, map, currentElement);
 		break;
+	case PGRN_WAL_ACTION_RENAME_TABLE:
+		PGrnWALApplyRenameTable(data, map, currentElement);
+		break;
+	case PGRN_WAL_ACTION_DELETE:
+		PGrnWALApplyDelete(data, map, currentElement);
+		break;
 	default:
 		ereport(ERROR,
 				(errcode(ERRCODE_INVALID_PARAMETER_VALUE),
@@ -1513,9 +1835,10 @@ PGrnWALApplyObject(PGrnWALApplyData *data, msgpack_object *object)
 	}
 }
 
-static void
+static int64_t
 PGrnWALApplyConsume(PGrnWALApplyData *data)
 {
+	int64_t nAppliedOperations = 0;
 	BlockNumber i;
 	BlockNumber startBlock;
 	LocationIndex dataOffset;
@@ -1536,8 +1859,7 @@ PGrnWALApplyConsume(PGrnWALApplyData *data)
 		Page page;
 		size_t dataSize;
 
-		buffer = ReadBuffer(data->index, i);
-		LockBuffer(buffer, BUFFER_LOCK_SHARE);
+		buffer = PGrnWALReadLockedBuffer(data->index, i, BUFFER_LOCK_SHARE);
 		page = BufferGetPage(buffer);
 		dataSize = PGrnWALPageGetLastOffset(page) - dataOffset;
 		msgpack_unpacker_reserve_buffer(&unpacker, dataSize);
@@ -1545,6 +1867,8 @@ PGrnWALApplyConsume(PGrnWALApplyData *data)
 			   PGrnWALPageGetData(page) + dataOffset,
 			   dataSize);
 		UnlockReleaseBuffer(buffer);
+		if (dataSize == 0)
+			break;
 
 		msgpack_unpacker_buffer_consumed(&unpacker, dataSize);
 		while (MSGPACK_UNPACKER_NEXT(&unpacker, &unpacked))
@@ -1558,30 +1882,326 @@ PGrnWALApplyConsume(PGrnWALApplyData *data)
 			PGrnIndexStatusSetWALAppliedPosition(data->index,
 												 i,
 												 appliedOffset);
+			nAppliedOperations++;
 		}
 		dataOffset = 0;
 	}
 	msgpack_unpacked_destroy(&unpacked);
 	msgpack_unpacker_destroy(&unpacker);
+
+	return nAppliedOperations;
 }
 #endif
 
-void
+int64_t
 PGrnWALApply(Relation index)
 {
+	int64_t nAppliedOperations = 0;
 #ifdef PGRN_SUPPORT_WAL
 	PGrnWALApplyData data;
 
 	data.index = index;
 	if (!PGrnWALApplyNeeded(&data))
-		return;
+		return 0;
 
 	LockRelation(index, RowExclusiveLock);
 	PGrnIndexStatusGetWALAppliedPosition(data.index,
 										 &(data.current.block),
 										 &(data.current.offset));
 	data.sources = NULL;
-	PGrnWALApplyConsume(&data);
+	nAppliedOperations = PGrnWALApplyConsume(&data);
 	UnlockRelation(index, RowExclusiveLock);
 #endif
+	return nAppliedOperations;
+}
+
+/**
+ * pgroonga_wal_apply(indexName cstring) : bigint
+ */
+Datum
+pgroonga_wal_apply_index(PG_FUNCTION_ARGS)
+{
+	int64_t nAppliedOperations = 0;
+#ifdef PGRN_SUPPORT_WAL
+	Datum indexNameDatum = PG_GETARG_DATUM(0);
+	Datum indexOidDatum;
+	Oid indexOid;
+	Relation index;
+
+	if (!PGrnIsWritable())
+	{
+		ereport(ERROR,
+				(errcode(ERRCODE_E_R_E_MODIFYING_SQL_DATA_NOT_PERMITTED),
+				 errmsg("pgroonga: wal_apply: "
+						"can't apply WAL "
+						"while pgroonga.writable is false")));
+	}
+
+	indexOidDatum = DirectFunctionCall1(regclassin, indexNameDatum);
+	indexOid = DatumGetObjectId(indexOidDatum);
+	if (!OidIsValid(indexOid))
+	{
+		ereport(ERROR,
+				(errcode(ERRCODE_INVALID_NAME),
+				 errmsg("pgroonga: wal_apply: unknown index name: <%s>",
+						DatumGetCString(indexNameDatum))));
+	}
+
+	index = RelationIdGetRelation(indexOid);
+	PG_TRY();
+	{
+		if (!PGrnIndexIsPGroonga(index))
+		{
+			ereport(ERROR,
+					(errcode(ERRCODE_INVALID_NAME),
+					 errmsg("pgroonga: wal_apply: not PGroonga index: <%s>",
+							DatumGetCString(indexNameDatum))));
+		}
+		nAppliedOperations = PGrnWALApply(index);
+	}
+	PG_CATCH();
+	{
+		RelationClose(index);
+		PG_RE_THROW();
+	}
+	PG_END_TRY();
+	RelationClose(index);
+#else
+	ereport(ERROR,
+			(errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
+			 errmsg("pgroonga: WAL: not supported")));
+#endif
+	PG_RETURN_INT64(nAppliedOperations);
+}
+
+/**
+ * pgroonga_wal_apply() : bigint
+ */
+Datum
+pgroonga_wal_apply_all(PG_FUNCTION_ARGS)
+{
+	int64_t nAppliedOperations = 0;
+#ifdef PGRN_SUPPORT_WAL
+	LOCKMODE lock = AccessShareLock;
+	Relation indexes;
+	HeapScanDesc scan;
+	HeapTuple indexTuple;
+
+	if (!PGrnIsWritable())
+	{
+		ereport(ERROR,
+				(errcode(ERRCODE_E_R_E_MODIFYING_SQL_DATA_NOT_PERMITTED),
+				 errmsg("pgroonga: wal_apply: "
+						"can't apply WAL "
+						"while pgroonga.writable is false")));
+	}
+
+	indexes = heap_open(IndexRelationId, lock);
+	scan = heap_beginscan_catalog(indexes, 0, NULL);
+	while ((indexTuple = heap_getnext(scan, ForwardScanDirection)))
+	{
+		Form_pg_index indexForm = (Form_pg_index) GETSTRUCT(indexTuple);
+		Relation index;
+
+		if (!pg_class_ownercheck(indexForm->indexrelid, GetUserId()))
+			continue;
+
+		index = RelationIdGetRelation(indexForm->indexrelid);
+		if (!PGrnIndexIsPGroonga(index))
+		{
+			RelationClose(index);
+			continue;
+		}
+
+		PG_TRY();
+		{
+			nAppliedOperations += PGrnWALApply(index);
+		}
+		PG_CATCH();
+		{
+			RelationClose(index);
+			heap_endscan(scan);
+			heap_close(indexes, lock);
+			PG_RE_THROW();
+		}
+		PG_END_TRY();
+		RelationClose(index);
+	}
+	heap_endscan(scan);
+	heap_close(indexes, lock);
+#else
+	ereport(ERROR,
+			(errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
+			 errmsg("pgroonga: WAL: not supported")));
+#endif
+	PG_RETURN_INT64(nAppliedOperations);
+}
+
+#ifdef PGRN_SUPPORT_WAL
+static int64_t
+PGrnWALTruncate(Relation index)
+{
+	int64_t nTruncatedBlocks = 0;
+	BlockNumber i;
+	BlockNumber nBlocks;
+	GenericXLogState *state;
+
+	LockRelation(index, RowExclusiveLock);
+	nBlocks = RelationGetNumberOfBlocks(index);
+	if (nBlocks == 0)
+	{
+		UnlockRelation(index, RowExclusiveLock);
+		return 0;
+	}
+
+	state = GenericXLogStart(index);
+
+	{
+		Buffer buffer;
+		Page page;
+		PGrnWALMetaPageSpecial *pageSpecial;
+
+		buffer = PGrnWALReadLockedBuffer(index,
+										 PGRN_WAL_META_PAGE_BLOCK_NUMBER,
+										 BUFFER_LOCK_EXCLUSIVE);
+		page = GenericXLogRegisterBuffer(state, buffer, GENERIC_XLOG_FULL_IMAGE);
+		pageSpecial = (PGrnWALMetaPageSpecial *) PageGetSpecialPointer(page);
+		pageSpecial->next = PGRN_WAL_META_PAGE_BLOCK_NUMBER + 1;
+		UnlockReleaseBuffer(buffer);
+
+		nTruncatedBlocks++;
+	}
+
+	for (i = PGRN_WAL_META_PAGE_BLOCK_NUMBER + 1; i < nBlocks; i++)
+	{
+		Buffer buffer;
+		Page page;
+
+		buffer = PGrnWALReadLockedBuffer(index, i, BUFFER_LOCK_EXCLUSIVE);
+		if (nTruncatedBlocks >= MAX_GENERIC_XLOG_PAGES)
+		{
+			GenericXLogFinish(state);
+			state = GenericXLogStart(index);
+		}
+		page = GenericXLogRegisterBuffer(state, buffer, GENERIC_XLOG_FULL_IMAGE);
+		PageInit(page, BLCKSZ, 0);
+		UnlockReleaseBuffer(buffer);
+
+		nTruncatedBlocks++;
+	}
+
+	GenericXLogFinish(state);
+
+	UnlockRelation(index, RowExclusiveLock);
+
+	PGrnIndexStatusSetWALAppliedPosition(index,
+										 PGRN_WAL_META_PAGE_BLOCK_NUMBER + 1,
+										 0);
+
+	return nTruncatedBlocks;
+}
+#endif
+
+/**
+ * pgroonga_wal_truncate(indexName cstring) : bigint
+ */
+Datum
+pgroonga_wal_truncate_index(PG_FUNCTION_ARGS)
+{
+	int64_t nTruncatedBlocks = 0;
+#ifdef PGRN_SUPPORT_WAL
+	Datum indexNameDatum = PG_GETARG_DATUM(0);
+	Datum indexOidDatum;
+	Oid indexOid;
+	Relation index;
+
+	indexOidDatum = DirectFunctionCall1(regclassin, indexNameDatum);
+	indexOid = DatumGetObjectId(indexOidDatum);
+	if (!OidIsValid(indexOid))
+	{
+		ereport(ERROR,
+				(errcode(ERRCODE_INVALID_NAME),
+				 errmsg("pgroonga: wal_truncate: unknown index name: <%s>",
+						DatumGetCString(indexNameDatum))));
+	}
+
+	index = RelationIdGetRelation(indexOid);
+	PG_TRY();
+	{
+		if (!PGrnIndexIsPGroonga(index))
+		{
+			ereport(ERROR,
+					(errcode(ERRCODE_INVALID_NAME),
+					 errmsg("pgroonga: wal_truncate: not PGroonga index: <%s>",
+							DatumGetCString(indexNameDatum))));
+		}
+		nTruncatedBlocks = PGrnWALTruncate(index);
+	}
+	PG_CATCH();
+	{
+		RelationClose(index);
+		PG_RE_THROW();
+	}
+	PG_END_TRY();
+	RelationClose(index);
+#else
+	ereport(ERROR,
+			(errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
+			 errmsg("pgroonga: WAL: not supported")));
+#endif
+	PG_RETURN_INT64(nTruncatedBlocks);
+}
+
+/**
+ * pgroonga_wal_truncate() : bigint
+ */
+Datum
+pgroonga_wal_truncate_all(PG_FUNCTION_ARGS)
+{
+	int64_t nTruncatedBlocks = 0;
+#ifdef PGRN_SUPPORT_WAL
+	LOCKMODE lock = AccessShareLock;
+	Relation indexes;
+	HeapScanDesc scan;
+	HeapTuple indexTuple;
+
+	indexes = heap_open(IndexRelationId, lock);
+	scan = heap_beginscan_catalog(indexes, 0, NULL);
+	while ((indexTuple = heap_getnext(scan, ForwardScanDirection)))
+	{
+		Form_pg_index indexForm = (Form_pg_index) GETSTRUCT(indexTuple);
+		Relation index;
+
+		if (!pg_class_ownercheck(indexForm->indexrelid, GetUserId()))
+			continue;
+
+		index = RelationIdGetRelation(indexForm->indexrelid);
+		if (!PGrnIndexIsPGroonga(index))
+		{
+			RelationClose(index);
+			continue;
+		}
+
+		PG_TRY();
+		{
+			nTruncatedBlocks += PGrnWALTruncate(index);
+		}
+		PG_CATCH();
+		{
+			RelationClose(index);
+			heap_endscan(scan);
+			heap_close(indexes, lock);
+			PG_RE_THROW();
+		}
+		PG_END_TRY();
+		RelationClose(index);
+	}
+	heap_endscan(scan);
+	heap_close(indexes, lock);
+#else
+	ereport(ERROR,
+			(errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
+			 errmsg("pgroonga: WAL: not supported")));
+#endif
+	PG_RETURN_INT64(nTruncatedBlocks);
 }
